/*!
 * Copyright (C) 2006 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Log$
 */

/*
 * Limited AT91SAM9260-EK MP3 player sample.
 *
 * Will sequentially play all MP3 files from an MMC's root directory.
 *
 * MultiMedia Cards only, no SD Cards (see arch/arm/dev/at91_mmc.c).
 * Root directory must contain MP3 files only.
 * Hardcoded for 180MHz CPU clock and 44.1kHz sampling rate (also see ./at73dac.c).
 */

#include <cfg/os.h>
#include <cfg/clock.h>
#include <dev/board.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <io.h>

#include <sys/version.h>
#include <sys/confnet.h>
#include <sys/atom.h>
#include <sys/heap.h>
#include <sys/thread.h>
#include <sys/timer.h>
#include <sys/event.h>
#include <sys/socket.h>

#include <hxmp3/mp3dec.h>

#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <net/route.h>

#include <pro/dhcp.h>
#include <pro/sntp.h>
#include <pro/discover.h>

#include "at73dac.h"
#include "config.h"
#include "shoutcast.h"
#include "player.h"
#include "httpserv.h"

/*
 * Unique MAC address of the Ethernut Board.
 *
 * Ignored if EEPROM contains a valid configuration.
 */
#define MY_MAC { 0x00, 0x06, 0x98, 0x30, 0x00, 0x35 }

/*
 * Unique IP address of the Ethernut Board.
 *
 * Ignored if DHCP is used.
 */
#define MY_IPADDR "192.168.192.35"

/*
 * IP network mask of the Ethernut Board.
 *
 * Ignored if DHCP is used.
 */
#define MY_IPMASK "255.255.255.0"

/*
 * Gateway IP address for the Ethernut Board.
 *
 * Ignored if DHCP is used.
 */
#define MY_IPGATE "192.168.192.1"

/*
 * Wether we should run a web interface.
 */
#if defined(AT91SAM9260_EK)
#define USE_HTTPSERVER
#endif

/*
 * Wether we should run a discovery responder.
 */
#if defined(AT91SAM9260_EK)
#define USE_DISCOVERY
#endif

/*
 * IP address of the host running a time daemon. 
 */
#if defined(AT91SAM9260_EK)
//#define MY_TIMED "130.149.17.21"
//#define MY_TIMED "192.168.192.223"
#endif

/*
 * Local timezone, -1 for Central Europe. 
 */
#ifndef MY_TZ
#define MY_TZ    -1
#endif

/*
 * Query a time server and optionally update the hardware clock.
 */
static int QueryTimeServer(void)
{
    int rc = -1;

#ifdef MY_TIMED
    int retries = 5;

    _timezone = MY_TZ * 60L * 60L;
    while (retries--) {
        time_t now;
        u_long timeserver = inet_addr(MY_TIMED);

        /* Query network time service and set the system time. */
        printf("Query time from %s...", MY_TIMED);
        if(NutSNTPGetTime(&timeserver, &now) == 0) {
            puts("OK");
            rc = 0;
            stime(&now);
            break;
        }
        else {
            puts("failed");
        }
    }
#endif

    return rc;
}

int InitEthernetDevice(void)
{
    u_long ip_addr = inet_addr(MY_IPADDR);
    u_long ip_mask = inet_addr(MY_IPMASK);
    u_long ip_gate = inet_addr(MY_IPGATE);
    u_char mac[6] = MY_MAC;

    for (;;) {
        printf("Register %s...", DEV_ETHER_NAME);
        if (NutRegisterDevice(&DEV_ETHER, 0x8300, 5) == 0) {
            puts("OK");
            break;
        }
        puts("No Ethernet, retry");
    }

    printf("Configure %s...", DEV_ETHER_NAME);
    if (NutDhcpIfConfig(DEV_ETHER_NAME, 0, 60000) == 0) {
        puts("OK");
        return 0;
    }
    printf("initial boot...");
    if (NutDhcpIfConfig(DEV_ETHER_NAME, mac, 60000) == 0) {
        puts("OK");
        return 0;
    }
    printf("No DHCP...");
    NutNetIfConfig(DEV_ETHER_NAME, mac, ip_addr, ip_mask);
    /* Without DHCP we had to set the default gateway manually.*/
    if(ip_gate) {
        printf("hard coded gate...");
        NutIpRouteAdd(0, 0, ip_gate, &DEV_ETHER);
    }
    puts("OK");

    return 0;
}

/*
 * Main application routine.
 */
int main(void)
{
    u_long baud = 115200;
    STATIONINFO *sip;
    RECEIVERINFO *rip_sc;
    PLAYERINFO *pip_mp3;
    u_long ipgate;
    int idx;

    /*
     * Register and open the DBGU device, which we will use for debug 
     * output via stdout.
     */
    NutRegisterDevice(&DEV_DEBUG, 0, 0);
    freopen(DEV_DEBUG_NAME, "w", stdout);
    _ioctl(_fileno(stdout), UART_SETSPEED, &baud);
    printf("\n\nSAM Internet Radio %s - Nut/OS %s\n", VERSION, NutVersionString());
    printf("%lu bytes free\n", NutHeapAvailable());

    printf("CPU Clock   : %lu\n", NutGetCpuClock());
    printf("Master Clock: %lu\n", At91GetMasterClock());
    printf("%lu system ticks/s\n", NutGetTickClock());

    /*
     * Adapt the CPU clock to the sample rate. Note, that we must also
     * set the baudrate for the new clock.
     */
#if 1
    if (At73DacAdaptMasterClock(44100)) {
        printf("Failed to set DAC clocks.\n");
        for (;;);
    }
    _ioctl(_fileno(stdout), UART_SETSPEED, &baud);
    printf("CPU Clock   : %lu\n", NutGetCpuClock());
    printf("Master Clock: %lu\n", At91GetMasterClock());
    printf("%lu system ticks/s\n", NutGetTickClock());
#endif

    /* 
     * Initialize the Ethernet device and print our IP setting. 
     */
    InitEthernetDevice();
    printf("IP Addr: %s\n", inet_ntoa(confnet.cdn_ip_addr));
    printf("IP Mask: %s\n", inet_ntoa(confnet.cdn_ip_mask));
    NutIpRouteQuery(0, &ipgate);
    printf("IP Gate: %s\n", inet_ntoa(ipgate));

    /* 
     * Register a discovery responder. This optional feature
     * allows to discover all Nut/OS devices in a local network.
     */
#ifdef USE_DISCOVERY
    printf("Start Responder...");
    if (NutRegisterDiscovery((u_long)-1, 0, DISF_INITAL_ANN)) {
        puts("failed");
    }
    else {
        puts("OK");
    }
#endif

    /*
     * Try to get current date and time from an SNTP server.
     */
    if (QueryTimeServer() == 0) {
        time_t now = time(0);
        struct _tm *lot = localtime(&now);
        printf("Date: %02u.%02u.%u\n", lot->tm_mday, lot->tm_mon + 1, 1900 + lot->tm_year);
        printf("Time: %02u:%02u:%02u\n", lot->tm_hour, lot->tm_min, lot->tm_sec);
#ifdef USE_HTTPSERVER
        h_timevalid = 1;
#endif
    }

#ifdef USE_HTTPSERVER
    HttpServerStart();
#endif

#if 0
    outr(PIOA_PUER, _BV(31));
    outr(PIOA_PER, _BV(31));
    outr(PIOA_ODR, _BV(31));
#endif

    /* Slow reset, the extension board needs time. */
    NutSleep(300);
    At73DacInit();

    ConfigResetFactory();

    rip_sc = ReceiverCreate(RECEIVER_TYPE_SHOUTCAST);
    pip_mp3 = PlayerCreate(PLAYER_TYPE_MP3);

    for (;;) {
        printf("Starting list\n");
        for (idx = 0; idx < MAXNUM_STATIONS; idx++) {
            if (station[idx].rs_port) {
                printf("Trying station entry %d\n", idx);
                sip = StationConnect(&station[idx]);
                if (sip) {
                    printf("Connected: %lu bytes free\n", NutHeapAvailable());
#ifdef USE_HTTPSERVER
                    HttpSetStationInfo(idx, sip);
#endif
                    if (sip->si_stype == RECEIVER_TYPE_SHOUTCAST) {
                        if (ReceiverStart(rip_sc, sip) == 0) {
#ifdef USE_HTTPSERVER
                            HttpSetReceiverInfo(rip_sc);
#endif
                            if (sip->si_ptype == PLAYER_TYPE_MP3) {
                                if (PlayerStart(pip_mp3, rip_sc) == 0) {
#ifdef USE_HTTPSERVER
                                    HttpSetPlayerInfo(pip_mp3);
#endif
                                    NutSleep(10000);
                                    for (;;) {
                                        NutSleep(1000);
                                        if (ReceiverThreadStatus(rip_sc) != RSTAT_RUNNING) {
                                            break;
                                        }
                                        if (PlayerThreadStatus(pip_mp3) != PSTAT_RUNNING) {
                                            break;
                                        }
    #if 0
                                        if ((inr(PIOA_PDSR) & _BV(31)) == 0) {
                                            break;
                                        }
    #endif
                                    }
                                    printf("\nStop player ");
#ifdef USE_HTTPSERVER
                                    HttpSetPlayerInfo(NULL);
#endif
                                    PlayerStop(pip_mp3);
                                }
                            }
                            else {
                                printf("Bad content type\n");
                            }
                            printf("\nStop receiver ");
#ifdef USE_HTTPSERVER
                            HttpSetReceiverInfo(NULL);
#endif
                            ReceiverStop(rip_sc);
                        }
                    }
                    else {
                        printf("Bad station type\n");
                    }
                    printf("\nDisconnect station ");
#ifdef USE_HTTPSERVER
                    HttpSetStationInfo(0, NULL);
#endif
                    StationDisconnect(sip);
                    printf(" Sleeping\n");
                    NutSleep(5000);
                }
                NutSleep(1000);
            }
        }
    }
    printf("End: %lu bytes free\n", NutHeapAvailable());
    for(;;) {
        putchar('.');
        NutSleep(1000);
    }
}
